/**
 * @file Graphics.c
 * Graphics driver ported from Chip Gracy's Graphics.spin
 * Users should comment out unused functions to save code space if necessary.
 * Copyright (c) 2009, Steve Denson
 * See end of file for MIT license terms.
 */
#include <stdio.h>
#include <string.h>
#include <propeller.h>
#include "Graphics.h"

/**
''
'' Theory of Operation:
''
'' A cog is launched with code contained in Graphics_Array[] which
'' processes commands sent by the Graph_* API wrapper functions.
''
'' Points, lines, arcs, sprites, text, and polygons are rasterized into
'' a specified stretch of memory which serves as a generic bitmap buffer.
''
'' The bitmap can be displayed by the TV or VGA driver.
'' To date, the C library version has been tested on TV only.
''
'' See GraphicsTest.c for usage example.
''
 */ 

/**
 * Graphics_Array is the PASM Graphics engine code.
 */ 
extern long Graphics_Array[];
extern short Graphics_Font[];

/**
 * Graphics interface control structure
 */
static GraphCmd_ST gGraph;

static gr_uint32_t pa[GRAPH_MAX_ARGS]; // set command parameter array

/**
 * Color codes
 */
gr_uint32_t Graph_colors[] =
{
  0x00000000,
  0x55555555,
  0xAAAAAAAA,
  0xFFFFFFFF
};

/**
 * Round pixel recipes
 */
static char Graph_pixels[] =
{
    0x00, 0x00, 0x00, 0x00,         //0,1,2,3
    0x00, 0x00, 0x02, 0x05,         //4,5,6,7
    0x0A, 0x0A, 0x1A, 0x1A,         //8,9,A,B
    0x34, 0x3a, 0x74, 0x74          //C,D,E,F
};

/*
 * Vector font in separate GraphicsFont.c file
 */


/**
 * start graphics driver
 * see header for API description.
 */
int  Graph_start(void)
{
    // spin uses fontptr := font. C has no such luxury
    // to achieve the same result, we pass font in setup
    int wt=2000000,cog;
    Graph_stop();
    cog = cognew_native((void*)Graphics_Array, (gr_uint32_t*)&gGraph.command)+1;
    gGraph.cog = cog;
    WAIT(wt);
    return cog;
}

/**
 * stop graphics driver
 * see header for API description.
 */
void Graph_stop(void)
{
    GraphCmd_ST* gp = &gGraph;
    gr_uint32_t cog = gp->cog-1;
    if(gp->cog > 0)
        COGSTOP(cog);
    gp->cog = 0;
}

/**
 * set bitmap parameters
 * see header for API description.
 */
void Graph_setup(gr_uint32_t x_tiles, gr_uint32_t y_tiles, gr_uint32_t x_origin, gr_uint32_t y_origin, gr_uint32_t base_ptr)
{
    GraphCmd_ST* gp = &gGraph;
    int n = 0;

    sendCommand(GraphCmd_loop, 0);

    for(n = 0; n < x_tiles && n < 32; n++)
        gp->bases[n] = base_ptr + ((n * y_tiles) << 6);

    y_tiles <<= 4;
    y_origin = y_tiles - y_origin - 1;

    pa[0] = x_tiles;
    pa[1] = y_tiles;
    pa[2] = x_origin;
    pa[3] = y_origin;
    pa[4] = (gr_uint32_t)gp->bases; // driver ignores arg4
    pa[5] = (gr_uint32_t)gp->bases;
    pa[6] = (gr_uint32_t)gp->slices;
    pa[7] = (gr_uint32_t)Graphics_Font;
    sendCommand(GraphCmd_setup, pa);

    gp->bitmap_ptr = (gr_uint32_t*)base_ptr;
    gp->bitmap_len = x_tiles * y_tiles;
    //printf("\r\nBP Longs 0x%x Base 0x%x\n", gp->bitmap_longs, gp->bitmap_base);
}

/**
 * clear bitmap
 * see header for API description.
 */
void Graph_clear(void)
{
    sendCommand(GraphCmd_loop, 0);
    memset(gGraph.bitmap_ptr, 0, gGraph.bitmap_len<<2);
}

/**
 * copy bitmap - use for double-buffered display
 * see header for API description.
 */
void Graph_copy(gr_uint32_t dest_ptr)
{
    sendCommand(GraphCmd_loop, 0);
    memcpy((void*)dest_ptr, (void*)gGraph.bitmap_ptr, gGraph.bitmap_len<<2);
}

/**
 * set pixel color to two bit pattern
 * see header for API description.
 */
void Graph_color(gr_uint32_t c)
{
    sendCommand(GraphCmd_color, (gr_uint32_t*)&Graph_colors[c & 3]);
}

/**
 * set pixel width - actual width is w[3..0]+1
 * see header for API description.
 */
void Graph_width(gr_uint32_t w)
{
    int  passes, r, n, pw;

    r = ~w & 0x10;
    w &= 0xf;
    gGraph.pixel_width = w;
    passes = (w >> 1) + 1;

    memset(pa, 0, GRAPH_MAX_ARGS<<2);
    pa[0] = w;
    pa[1] = passes;

    sendCommand(GraphCmd_width, pa);

    pw = w ^ 0xf;
    for(n = 0; n <= w >> 1; n++) {
        gGraph.slices[n] = (gr_uint32_t)(0xffffffff >> (pw << 1)) << (pw & 0xE);
        if(r && (Graph_pixels[w] & (1 << n)))
            pw += 2;
        if(r && n == passes - 2)
            pw += 2;
    }
}

/**
 * set pixel color and width
 * see header for API description.
 */
void Graph_colorwidth(gr_uint32_t c, gr_uint32_t w)
{
    Graph_color(c);
    Graph_width(w);
}

/**
 * plot a pixel at point x,y
 * see header for API description.
 */
void Graph_plot(gr_uint32_t x, gr_uint32_t y)
{
    pa[0] = x;
    pa[1] = y;
    sendCommand(GraphCmd_plot, pa);
}

/**
 * draw a line to point x,y (from previous x,y?)
 * see header for API description.
 */
void Graph_line(gr_uint32_t x, gr_uint32_t y)
{
    pa[0] = x;
    pa[1] = y;
    sendCommand(GraphCmd_line, pa);
}

/**
 * draw an arc
 * see header for API description.
 */
void Graph_arc(gr_uint32_t x, gr_uint32_t y, gr_uint32_t xr, gr_uint32_t yr, int angle, int anglestep, int steps, int arcmode)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = xr;
    pa[3] = yr;
    pa[4] = angle;
    pa[5] = anglestep;
    pa[6] = steps;
    pa[7] = arcmode;
    sendCommand(GraphCmd_arc, pa);
}

/**
 * draw a vector
 * see header for API description.
 */
void Graph_vec(gr_uint32_t x, gr_uint32_t y, int vecscale, int vecangle, gr_uint16_t* vecdef_ptr)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = vecscale;
    pa[3] = vecangle;
    pa[4] = (gr_uint32_t)vecdef_ptr;
    sendCommand(GraphCmd_vec, pa);
}

/**
 * draw a vector arc
 * see header for API description.
 */
void Graph_vecarc(gr_uint32_t x, gr_uint32_t y, gr_uint32_t xr, gr_uint32_t yr, int angle, int vecscale, int vecangle, gr_uint16_t* vecdef_ptr)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = xr;
    pa[3] = yr;
    pa[4] = angle;
    pa[5] = vecscale;
    pa[6] = vecangle;
    pa[7] = (gr_uint32_t)vecdef_ptr;
    sendCommand(GraphCmd_vecarc, pa);
}

/**
 * draw a pixel sprite
 * see header for API description.
 */
void Graph_pix(gr_uint32_t x, gr_uint32_t y, int pixrot, gr_uint16_t* pixdef_ptr)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = pixrot;
    pa[3] = (long)pixdef_ptr;
    sendCommand(GraphCmd_pix, pa);
}

/**
 * draw a pixel sprite at an arc position
 * see header for API description.
 */
void Graph_pixarc(gr_uint32_t x, gr_uint32_t y, gr_uint32_t xr, gr_uint32_t yr, int angle, int pixrot, gr_uint16_t* pixdef_ptr)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = xr;
    pa[3] = yr;
    pa[4] = angle;
    pa[5] = pixrot;
    pa[6] = (gr_uint32_t)pixdef_ptr;
    sendCommand(GraphCmd_pixarc, pa);
}

/**
 * draw char at x,y
 * see header for API description.
 */
void Graph_char(gr_uint32_t x, gr_uint32_t y, char ch)
{
    gr_uint32_t just[2];

    char s[2] = { '\0', '\0' };
    s[0] = ch;
    Graph_text(x, y, s);
}

/**
 * draw text at x,y
 * see header for API description.
 */
void Graph_text(gr_uint32_t x, gr_uint32_t y, char* string_ptr)
{
    gr_uint32_t just[2];

    justify(string_ptr, just);
    pa[0] = x;
    pa[1] = y;
    pa[2] = (gr_uint32_t)string_ptr;
    pa[3] = just[0];
    pa[4] = just[1];
    sendCommand(GraphCmd_text, pa);
}

/**
 * draw text at an arc position
 * see header for API description.
 */
void Graph_textarc(gr_uint32_t x, gr_uint32_t y, gr_uint32_t xr, gr_uint32_t yr, int angle, char* string_ptr)
{
    gr_uint32_t just[2];

    justify(string_ptr, just);
    pa[0] = x;
    pa[1] = y;
    pa[2] = xr;
    pa[3] = yr;
    pa[4] = angle;
    pa[5] = (gr_uint32_t)string_ptr;
    pa[6] = just[0];
    pa[7] = just[1];
    sendCommand(GraphCmd_textarc, pa);
}

/**
 * set text size and justification
 * see header for API description.
 */
void Graph_textmode(gr_uint32_t x_scale, gr_uint32_t y_scale, gr_uint32_t spacing, gr_uint32_t justification)
{
    GraphCmd_ST* gp = &gGraph;
    pa[0] = gp->text_xs = x_scale;
    pa[1] = gp->text_ys = y_scale;
    pa[2] = gp->text_sp = spacing;
    pa[3] = gp->text_just = justification;
    sendCommand(GraphCmd_textmode, pa);
}

/**
 * draw a box with round/square corners according to pixel width
 * see header for API description.
 */
void Graph_box(int x, int y, int box_width, int box_height)
{
    gr_uint32_t x2, y2, pmin, pmax;
    int width = gGraph.pixel_width;

    if((box_width > width) && (box_height > width))
    {
        pmin = width >> 1;
        pmax = width - pmin;
        x += pmin;
        y += pmin;
        x2 = x + box_width  - 1 - width;
        y2 = y + box_height - 1 - width;

        Graph_plot(x,y);
        Graph_plot(x,y2);
        Graph_plot(x2,y);
        Graph_plot(x2,y2);

        fill(x, y2 + pmax, 0, (x2 - x) << 16, 0, 0, pmax);  // fill gaps
        fill(x, y, 0, (x2 - x) << 16, 0, 0, pmin);
        fill(x - pmin, y2, 0, (x2 - x + width) << 16, 0, 0, y2 - y);
    }
}

/**
 * draw a solid quadrilateral
 * see header for API description.
 */
void Graph_quad(gr_uint32_t x1, gr_uint32_t y1, gr_uint32_t x2, gr_uint32_t y2, gr_uint32_t x3, gr_uint32_t y3, gr_uint32_t x4, gr_uint32_t y4)
{
    Graph_tri(x1, y1, x2, y2, x3, y3);
    Graph_tri(x3, y3, x4, y4, x1, y1);
}

void Graph_tri(gr_uint32_t x1, gr_uint32_t y1, gr_uint32_t x2, gr_uint32_t y2, gr_uint32_t x3, gr_uint32_t y3)
{
    int xy[2];
    int mycase = ((y1 >= y2) ? 4 : 0) | ((y2 >= y3) ? 2 : 0) | ((y1 >= y3) ? 1 : 0);

    switch(mycase) {
        case 0:
            xy[0] = x1; 
            xy[1] = y1; // 2 longs 
            x1 = x3;
            y1 = y3;    // 2 longs
            x3 = xy[0];
            y3 = xy[1]; // 2 longs
            break;
        case 2:
            xy[0] = x1; 
            xy[1] = y1; // 2 longs
            x1 = x2;
            y1 = y2;
            x2 = x3;
            y2 = y3;    // 4 longs
            x3 = xy[0];
            y3 = xy[1]; // 2 longs
            break;
        case 3:
            xy[0] = x1; 
            xy[1] = y1; // 2 longs
            x1 = x2;
            y1 = y2;    // 2 longs
            x2 = xy[0];
            y2 = xy[1]; // 2 longs
            break;
        case 4:
            xy[0] = x3; 
            xy[1] = y3; // 2 longs
            x2 = x1;
            y2 = y1;
            x3 = x2;
            y3 = y2;    // 4 longs
            x1 = xy[0];
            y1 = xy[1]; // 2 longs
            break;
        case 5:
            xy[0] = x2; 
            xy[1] = y2; // 2 longs
            x2 = x3;
            y2 = y3;    // 2 longs
            x3 = xy[0];
            y3 = xy[1]; // 2 longs
            break;
        default:
            break;
    }

    fill(x1, y1, ((x3 - x1) << 16) / (y1 - y3 + 1), ((x2 - x1) << 16) / (y1 - y2 + 1), ((x3 - x2) << 16) / (y2 - y3 + 1), y1 - y2, y1 - y3);
}

/**
 * make sure last command is done
 */
void Graph_finish(void)
{
    sendCommand(GraphCmd_loop, 0);
}


/**
 * fill blocks
 */
static void fill(int x, int y, int da, int db, int db2, int linechange, int lines_minus_1)
{
    pa[0] = x;
    pa[1] = y;
    pa[2] = da;
    pa[3] = db;
    pa[4] = db2;
    pa[5] = linechange;
    pa[6] = lines_minus_1;
    sendCommand(GraphCmd_fill, pa);
}

/**
 * justify apparently aligns string edges
 */
static void justify(char* string_ptr, gr_uint32_t* justptr)
{
    GraphCmd_ST* gp = &gGraph;
    gr_uint32_t x = (strlen(string_ptr) - 1) * gp->text_xs * gp->text_sp + gp->text_xs * 5 - 1;
    gr_uint32_t ja1[4];
    gr_uint32_t ja2[4];
    
    // p1 is an alias of ja1
    ja1[0] = 0;
    ja1[1] = x >> 1;
    ja1[2] = x;
    ja1[3] = 0;
    justptr[0] = -ja1[gp->text_just]; // -lookupz(text_just >> 2 & 3: 0, x >> 1, x, 0)

    // p2 is an alias of ja1
    ja2[0] = 0;
    ja2[1] = gp->text_ys << 3;
    ja2[2] = gp->text_ys << 4;
    ja2[3] = 0;
    justptr[1] = -ja2[gp->text_just]; // -lookupz(text_just & 3: 0, text_ys << 3, text_ys << 4, 0)
}

//#include <asio.h>
#ifdef __ASIO_H__
char* cmds[] = {
    "None",  "Setup",
    "Color", "Width",
    "Plot",  "Line", 
    "Arc",   "Vec", "VecArc",
    "Pix",   "PixArc",
    "Text",  "TextArc", "TextMode",
    "Fill",  "Loop",
    0 };
#endif

/**
 * send command to driver
 * @param cmd    - GraphCmd_ET command
 * @param argptr - arguments pointer
 */
static void sendCommand(GraphCmd_ET cmd, gr_uint32_t* argptr)
{
    //write command and pointer
    gGraph.command = ((cmd << 16) | ((int)argptr & 0xffff));
#ifdef __ASIO_H__
    printf("\r\nsendCommand(0x%x, 0x%x): 0x%x ... %s", cmd, (int)argptr, command, cmds[cmd]);
#endif

    // wait for command to be cleared, signifying receipt
    while(gGraph.command != 0)
        ;
    //OUTA ^= 4;
}



/*
 * Ported from Graphics.spin
 *
''***************************************
''*  Graphics Driver v1.0               *
''*  Author: Chip Gracey                *
''*  Copyright (c) 2005 Parallax, Inc.  *               
''*  See end of file for terms of use.  *               
''***************************************
 * 
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                  
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/
