/**
 * @file mouse.h
 * ICC Propeller mouse driver API
 *
 * Copyright (c) 2008, Steve Denson
 * See end of file for terms of use.
 */
#ifndef __MOUSE__
#define __MOUSE__

/**
 * mouse pin definitions for protoboard
 */
#define MOUSE_PROTO_DATA_PIN  24
#define MOUSE_PROTO_CLOCK_PIN 25
 
/**
 * mouse parameter table struct
 */
typedef struct mouse_struct
{
    long  x;        // absolute x     read-only  (7 contiguous longs)
    long  y;        // absolute y     read-only
    long  z;        // absolute z     read-only
    long  buttons;  // button states  read-only
    long  present;  // mouse present  read-only
    long  dpin;     // data pin       write-only
    long  cpin;     // clock pin      write-only
} mouse_t;

/**
 * mouse type enum
 *  3 = five-button scrollwheel mouse
 *  2 = three-button scrollwheel mouse
 *  1 = two-button or three-button mouse
 *  0 = not present
 */
typedef enum {
    MOUSE_NOT_PRESENT,
    MOUSE_2OR3_BUTTON,
    MOUSE_3BUTTON_SCROLL_WHEEL,
    MOUSE_5BUTTON_SCROLL_WHEEL
} mouse_type_en;

/**
 * mouse buttons enum
 *   bit4 = right-side button
 *   bit3 = left-side button
 *   bit2 = center/scrollwheel button
 *   bit1 = right button
 *   bit0 = left button
 */
typedef enum {
    MOUSE_LEFT  = 0x1,
    MOUSE_RIGHT = 0x2,
    MOUSE_CENTER = 0x4,
    MOUSE_LEFT_SIDE = 0x8,
    MOUSE_RIGHT_SIDE = 0x10,
    MOUSE_BUTTON_MASK = 0x1f
} mouse_button_en;

/**
 * starts mouse driver
 * @param dpin is mouse data pin.
 * @param cpin is mouse clock pin.
 * @returns cogid (not cogid+1)
 */
int mouse_start(int dpin, int cpin);

/**
 * starts mouse driver
 * @param id is cog id returned by mouse_start
 */
void mouse_stop(int id);

/**
 * check if mouse is present
 * @returns mouse type if present, else 0
 */
mouse_type_en mouse_present(void);

/**
 * Get mouse buttons state that can be interpreted with mouse_button_en.
 * @returns mouse buttons state
 */
int mouse_buttons(void);

/**
 * Get mouse x position.
 * @returns mouse x
 */
int mouse_x(void);

/**
 * Get mouse y position.
 * @returns mouse y
 */
int mouse_y(void);

/**
 * Mouse deltas and bounds are omitted from this driver
 * as they wouldrequire globals and not be thread safe.
 * Such statistics should be kept by the user.
 */
 
/*
PUB stop
PUB present : type
PUB button(b) : state
PUB buttons : states
PUB abs_x : x
PUB abs_y : y
PUB abs_z : z
PUB delta_reset
PUB delta_x : x | newx
PUB delta_y : y | newy
PUB delta_z : z | newz
PUB bound_limits(xmin, ymin, zmin, xmax, ymax, zmax) | i
PUB bound_scales(x_scale, y_scale, z_scale)
PUB bound_preset(x, y, z) | i, d
PUB bound_x : x
PUB bound_y : y
PUB bound_z : z
PRI bound(i, delta) : b | d
*/

#endif
// __MOUSE__

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/