/**
 * @file sllist.c
 * This file defines the implementation for singly linked list
 *
 * Copyright (c) 2008, Steve Denson
 * See end of file for terms of use.
 */
#include "sllist.h"
#include <stdlib.h>

/**
 * The create function is called to create a handle to an sllist "object".
 * See header for API defs.
 */
sllist_t* sllist_create(void)
{
    sllist_t* slp = (sllist_t*) malloc(sizeof(sllist_t));
    slp->head = 0;
    slp->tail = 0;
    return slp;
}

/**
 * The destroy function is called to destroy the "object".
 * See header for API defs.
 */
void sllist_destroy(sllist_t* slp)
{
    sllist_node_t* cur  = 0;
    sllist_node_t* head = slp->head;
    while(head != 0) { 
        cur  = head;
	    head = head->next;
	    free(cur);
    }
    free(slp);
}

/**
 * The add function lets user add an entry to the list.
 * See header for API defs.
 */
int sllist_add(sllist_t* slp, void* ptr)
{
    sllist_node_t* tail = slp->tail;
    sllist_node_t* next = (sllist_node_t*) malloc(sizeof(sllist_node_t));
    if (!next)
        return 0;
    if (slp->tail)
        slp->tail->next = next;
    else
        slp->head = next;
    next->value = ptr;
    next->next = 0;
    slp->tail = next;
    slp->count++;
    return slp->count;
}

/**
 * The clear function is called to clear but not destroy the list.
 * See header for API defs.
 */
void sllist_clear(sllist_t* slp)
{
    sllist_node_t* cur = 0;
    sllist_node_t* head = slp->head;
    while(head != 0) { 
        cur = head;
	    head = head->next;
	    free(cur);
    }
}

/**
 * The delete function lets user delete an item from the list.
 * See header for API defs.
 */
int sllist_delete(sllist_t* slp, void* ptr)
{
    sllist_node_t* item = slp->head;
    sllist_node_t* temp = item;
    while(item) {
        if(item->next == ptr) {
	        temp = item->next;
	        item->next = temp->next;
	        free(ptr);
            slp->count--;
	        return slp->count;
	    }
	    item = item->next;
    }
    return 0;
}

/**
 * The get function lets user get item info by index from list.
 * See header for API defs.
 */
void* sllist_get(sllist_t* slp, int index)
{
    int ii = 1;
    sllist_node_t* item = slp->head;
    while(item) {
        if(index == ii)
	        return item->value;
	    ii++;
	    item = item->next;
    }
    return 0;
}

/**
 * The index function lets user get the index an item info list using ptr lookup.
 * See header for API defs.
 */
int sllist_index(sllist_t* slp, void* ptr)
{
    int ii = 1;
    sllist_node_t* item = slp->head;
    while(item) {
        if(item == ptr)
	        return ii;
	    ii++;
	    item = item->next;
    }
    return 0;
}

/**
 * The size function lets user get the size of the list.
 * See header for API defs.
 */
int sllist_size(sllist_t* slp)
{
    return slp->count;
}

#ifdef SLLIST_TEST
#include <stdio.h>
#include <vga_text.h>

int gmheap[256]; // malloc heap

/**
 * main test for singly linked list sllist module
 * - starts vga
 * - sets up heap for malloc
 * - creates and populates a list
 * - prints list
 * - destroys list
 */
void main(void)
{
    sllist_t* list;
    sllist_node_t* ptr;

    vgaText_start(16);
    _NewHeap(gmheap, gmheap+sizeof(gmheap)-1);

    list = sllist_create();

    sllist_add(list,"hello world\n");
    sllist_add(list,"how's your day?\n");
    sllist_add(list,"still spinning?\n");

    ptr = list->head;
    while(ptr) {
        printf(ptr->value);
        ptr = ptr->next;
    }
    sllist_destroy(list);
}
#endif

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/