//============================================================================
//  TSL230.c                      Allen Marincak            Ver 1.0   Mar 2009
//
//  TSL230R Light to Frequency sensor API, with manual or auto scaling.
//
//  Copyright (c) 2009, Allen Marincak       See end of file for terms of use.
//============================================================================
//
//  A port to C for the ImageCraft C compiler. Ported from tsl230.spin (c)
//  Paul Baker of Parallac Inc. Some discussions about scaling in the original
//  SPIN code were changed to more appropriately reflect the sensitivity 
//  options of the device ( scaling is hardwired to full scale and it is the
//  sesnitivity that the driver and API actually modify ). Also added a new
//  function to get both a sample and the sensitivity in effect at the time.
//  The original SetScale() function was renamed SetSensitivity() and was
//  modified to switch to manual mode or re-enable auto mode when called.
//
//============================================================================
//
//   Taos TSL230 light to frequency sensor connection diagram.
//
//                    +---+  +---+
//    ctrlpinbase ----|1  +--+  8|---+--- GND
//                    |          |   |  
//  ctrlpinbase+1 ----|2        7|---+ 
//                    |          |   
//                +---|3        6|--------- inpin
//                |   |          |    
//          GND --+---|4        5|---+-- 3.3V 
//                |   +----------+   |       
//                |                  |
//                +--------||--------+
//                      .1uF Ceramic Capacitor (recommended)
//
//
//  NOTES:  Pins 7 and 8 of the TSL230 are hard wired to ground. This gives
//          a divisor factor of 1 which is fine for the Prop (at 80 Mhz)
//          because it's counters are fast enough to handle the TSL230's
//          output frequencies at full speed (about 1 Mhz).
//
//          At 3.3vdc the output frequency is about 99.2 % of what it would 
//          be at 5.0 volts per the specs so when calculating irradiance be
//          aware that your readings will need to be adjusted with respect to
//          the published charts which use VCC = 5.0 V.
//
//============================================================================

#include <propeller.h>
#include "tsl230_array.h"

unsigned long   freq, cog;
unsigned long   sens0, sens1, sps, autosens;
unsigned long   ctrls0, ctrls1; 


//----------------------------------------------------------------------------
//  TSL230_Start()
//
//  p_in          - pin number output of tsl230 is connected
//  p_ctrlbase    - pin number connected to S0
//                  S1 connected to ctrlbase + 1
//  p_samplefreq  - number of measurements per second
//                  higher values reduce significant digits
//  p_autosens    - boolean indicates if autosensitivity is used
//                  - with autosens on, output range is between
//                    0 and ~160,000,000
//                  - with autosens off, output range is between
//                    0 and ~1,600,000
//
//  Returns     -1 on failure to start a COG
//              else the COG ID ( 0 - 7 )

//----------------------------------------------------------------------------
int TSL230_Start( unsigned long p_in, unsigned long p_ctrlbase, unsigned long p_samplefreq, unsigned long p_autosens )
{
    sens0 = 3;                        // set inital sensitivity to maximum
    ctrls0 = 1 << p_ctrlbase;               // copy parameters
    ctrls1 = 1 << (p_ctrlbase+1);
    sps = p_samplefreq;
    autosens = p_autosens;

    DIRA |= ctrls0;                         // set control pins to output
    DIRA |= ctrls1;
    OUTA |= ctrls0;                         // set sensitivity
    OUTA |= ctrls1;
      
    tsl230_array[11] = 0x28000000 + p_in;               // set counter mode and control pin
    tsl230_array[12] = 80000000 / p_samplefreq;         // set wait period
  
    cog = cognew_native( (void*)tsl230_array, &freq );  // start driver

    return( cog );
}


//----------------------------------------------------------------------------
//  TSL230_Stop()
//
//  Stop driver - frees a cog
//----------------------------------------------------------------------------
void TSL230_Stop( unsigned long cogid ) 
{
    COGSTOP( cogid );
}                                                               


//----------------------------------------------------------------------------
//  TSL230_GetSample()
//
//  Returns the current reading adjusted by the sensitivity to give values
//  normalized to the full size of the sensor's photodiode array (which is a 
//  sensitivity of 3, the high sensitivity setting). This will simplify 
//  calculating irradiance. Lowering the sensitivity effectively reduces the
//  photodiode area (just turns some off) thus the lower frequency values. The
//  result is at lower sensitivities there will be an apparent loss of 
//  significant digits in that the least significant digits will be zero.
//----------------------------------------------------------------------------
unsigned long TSL230_GetSample( void )
{
    unsigned long val;
    
    val = freq * sps;                           // compute the frequency
    if (sens0 == 1 ) val *= 100;
    if (sens0 == 2 ) val *= 10;
    
    sens1 = sens0;                              // save current sensitivity
    
    if ( autosens )                             // if autoscaling is on
    {
        if ( val > 1000000 )                    //  - if output > 1Mhz decrease gain
        {   
            sens0--;
            if ( sens0 < 1 ) sens0 = 1;
        }
        else if ( val < 10000 )                 // - if output < 10Khz increase gain
        {
            sens0++;
            if ( sens0 > 3 ) sens0 = 3;
        }
        
        if ( sens0 & 1 ) OUTA |= ctrls0;
        else OUTA &= ~ ctrls0;
        if ( sens0 & 2 ) OUTA |= ctrls1;
        else OUTA &= ~ ctrls1;        
    }
    
    return( val );
}
    

//----------------------------------------------------------------------------
//  TSL230_GetSampleSens()
//
//  returns current sample and sensitivity to pointers provided
//      psam = ptr to sample (frequency output by the sensor)
//      psen = ptr to sensitivity
//----------------------------------------------------------------------------
void TSL230_GetSampleSens( unsigned long *psam, unsigned long *psen )
{
    *psam = TSL230_GetSample();                 // get sample    
    *psen = sens1;                              // get sensitivity
}    

  
//----------------------------------------------------------------------------
//  TSL230_SetSensitivity()
//
//  SetScale manually sets the sensitivity and turns off the auto sensitivity.
//  A sensitivity of 0 will reenable auto senstivity mode. 
//
//  psens = sensitivity to set to, 1, 2, or 3 only
//          0 = switch on autol sensitivity
//          
//----------------------------------------------------------------------------
void TSL230_SetSensitivity( unsigned long psens )
{
    if ( psens )
    {
        autosens = 0;
        sens0 = psens;
        if ( sens0 > 3 ) sens0 = 3;
    
        if ( sens0 & 1 ) OUTA |= ctrls0;
        else OUTA &= ~ ctrls0;
        if ( sens0 & 2 ) OUTA |= ctrls1;
        else OUTA &= ~ ctrls1;
    }
    else
    {
        autosens = 1;
    }        
}

//============================================================================
//                        TERMS OF USE: MIT License                                                       
//============================================================================
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
//                                       
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
//============================================================================
