/**
 * @file timer.c
 *
 * This file provides a basic timer proc list implementation.
 * Only one timer proc list is supported for now.
 * Future implementation (with a parameterized cognew)
 * will allow multiple timers.
 *
 * Copyright (c) 2008, Steve Denson
 * See end of file for terms of use.
 */
#include "sllist.h"
#include "timer.h"
#include <stdlib.h>
#include <stdio.h>

#ifdef NOPROP
// NOPROP case is for building with GNU or MSVC++
void asm(char* s) {}
#else
// PROP case
#include <propeller.h>
#include <propclock.h>
#endif

/**
 * Global Variables.
 * Need to use a global timer_t* for now since we have no way to pass
 * parameters to cognew. When that ability exists, the module will
 * become more generic and the context can be given to the cognew function.
 * Until that time, this module will provide only one timer proc list.
 */
timer_t* gtimer = 0;
long timer_stack[100];

/**
 * The create function is called to create a timer handle.
 * See timer.h for API defs.
 */
timer_t* timer_create(int msec)
{
    timer_t* timer = 0;
    int ii = 0;
    timer = (timer_t*) malloc(sizeof(timer_t));
    if(timer) {
        timer_disable(timer);
        timer->id   = cognew(timer_proc, timer_stack);
        timer->msec = msec;
        timer->list = sllist_create();
        timer_enable(timer);
    }
    return timer;
}

/**
 * The destroy function is called to destroy the "object".
 * See timer.h for API defs.
 */
void   timer_destroy(timer_t* mp)
{
    int id = mp->id;
    if(mp->list)
        sllist_destroy(mp->list);
    mp->enabled = FALSE;
    free(mp);
    COGSTOP(id);
}

/**
 * The add function adds a callback function to the timer list.
 * See timer.h for API defs.
 */
int    timer_add(timer_t* mp, void* fptr)
{
    if(!mp || !fptr)
        return 0;
    if(!mp->list)
        return 0;
    return sllist_add(mp->list, fptr);
}

/**
 * The del function deletes a callback function from the timer list.
 * See timer.h for API defs.
 */
int    timer_delete(timer_t* mp, void* fptr)
{
    if(!mp || !fptr)
        return 0;
    return sllist_delete(mp->list, fptr);
}

/**
 * The status function gets the timer status.
 * See timer.h for API defs.
 */
int    timer_status(timer_t* mp)
{
    return mp->enabled ? 0 : 1;
}

/**
 * The enable function turns the timer on/off.
 * See timer.h for API defs.
 */
void   timer_enable(timer_t* mp)
{
    if(mp) mp->enabled = -1;
}

/**
 * The enable function turns the timer on/off.
 * See timer.h for API defs.
 */
void   timer_disable(timer_t* mp)
{
    if(mp) mp->enabled = 0;
}

/**
 * The proc function runs the functions defined in the timer list.
 */
void timer_proc(void)
{
    sllist_t* list = 0;
    sllist_node_t* item = 0;

    int time  = 0;
    int count = 0;
    
    while(1)
    {
        count = (gtimer->msec*CLKFREQ)/1000;
        time = CNT + count;

        if(!gtimer->enabled) {
            continue;
        }
        list = gtimer->list;
        item = list->head;
        while(item) {
            void (*fptr)(timer_t*) = (void*) item->value;
    	    (fptr)(gtimer);
    	    item = item->next;
        }
        if(time > CNT) {
            WAITCNT(time, count);
        }
    }
}

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/