/**
 * @file timer.h
 *
 * This file provides a timer interface. Each timer runs in a Propeller COG.
 * Up to N timers can be defined depending on the number of COGs available.
 * Timers are dynamically allocated from the COG list <TBD>.
 *
 * Copyright (c) 2008, Steve Denson
 * See end of file for terms of use.
 */
#ifndef __TIMER_H__
#define __TIMER_H__

#include "sllist.h"

/**
 * API data struct and enum declarations
 */

typedef enum {
  TIMER_COG_MAX = 8,
  TIMER_PROC_MAX = 4,
  TIMER_EN_LAST
} timer_en;


/**
 * The timer proc/function list is defined by the sllist module.
 * This typedef defines an acceptable callback function type.
 */
typedef void (*fptr)(void*);

/**
 * The timer struct defines the period of execution and
 * a list of functions to call in a timer period.
 * The function list is a singly linked list of function pointers.
 * Functions may be added or deleted.
 */
typedef struct timer_struct {
  int id;   // cog id
  int min;  // number of minutes to wait for next event
  int msec; // number of milli-seconds to wait for next event
  int usec; // number of micro-seconds to wait for next event
  int run;  // number of procs run in a period
  int count;// number of procs on the timer proc list
  int enabled;  // disable if non-zero, else enable
  sllist_t* list; // keep list last
} timer_t;

/**
 * API function signature declarations
 */
 
/**
 * The create function is called to create an handle to an timer "object".
 * The object contains all functions and data for an interface.
 * It is critical that you allow enough time for your timer process list to run.
 * Change the period if you like with timer_set(ptr, msec).
 * Disable timer before changing period and re-enable to ensure proper operation.
 * @param msec is number of milli-seconds for timer to run.
 * @returns timer_t object handle, or zero if no timer available.
 */
timer_t* timer_create(int msec);

/**
 * The destroy function is called to destroy the "object".
 * @param ptr is the object handle to destroy.
 */
void   timer_destroy(timer_t* ptr);

/**
 * The add function adds a callback function to the timer list.
 * @param fptr is the function pointer to add to the timer process list.
 * @returns non-zero if list add fails, else zero.
 */
int    timer_add(timer_t* ptr, void* fptr);

/**
 * The del function deletes a callback function from the timer list.
 * @param fptr is the function pointer to delete from the timer process list.
 * @returns non-zero if list delete fails, else zero.
 */
int    timer_delete(timer_t* ptr, void* fptr);

/**
 * The status function gets the timer.
 * If the timer expired before all functions on the proc list runs,
 * then the timer will be defunct.
 * @returns zero if ok, or non-zero if timer is defunct.
 */
int    timer_status(timer_t* ptr);

/**
 * The enable function turns the timer on.
 */
void   timer_enable(timer_t* ptr);

/**
 * The disable function turns the timer off.
 */
void   timer_disable(timer_t* ptr);

/**
 * The proc function runs the functions defined in the timer list.
 */
void timer_proc(void);

/**
 * Both user code and timer code needs this.
 */
extern timer_t* gtimer;

 
#endif
// __TIMER_H__

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/