//============================================================================
//  I2C.c                         Allen Marincak            Ver 1.0   Mar 2009
//
//      Mod and port to C from the I2C Spin Object V1.3 by James Burrows.
//
//  Copyright (c) 2009, Allen Marincak       See end of file for terms of use.
//============================================================================


#include <propeller.h>
#include "i2c.h"

#define _i2cACK     0
#define _i2cNAK     1

static void             i2cStop ( void );
static void             i2cStart( void );
static unsigned char    i2cWrite( unsigned char i2cData );
static unsigned char    i2cRead ( unsigned char ackbit);

unsigned long i2cDat, i2cClk;

//----------------------------------------------------------------------------
//  I2C_Init()
//
//      Store I2C pins and initialize them.
//----------------------------------------------------------------------------
void I2C_Init( unsigned char pinDat, unsigned char pinClk )
{
    i2cDat = 1 << pinDat;       // Propeller pin used for I2C Data
    i2cClk = 1 << pinClk;       // Propeller pin used for I2C Clock
    
    DIRA &= ~i2cDat;            // ensure Data Pin is input to start
    OUTA &= ~i2cDat;            //      - set Dat pin Output Latch to LOW    
    DIRA &= ~i2cClk;            // ensure Clock Pin is input to start
    OUTA &= ~i2cClk;            //      - set Clock pin Output Latch to LOW
}


//----------------------------------------------------------------------------
//  I2C_IsDevPresent()
//
//      Send i2c address to check for Ack bit.
//
//      returns 0 if no device at given address
//              1 if there is a device at the given address
//----------------------------------------------------------------------------
unsigned char I2C_IsDevPresent( unsigned char devAddr )
{
    unsigned char ackbit;

    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    i2cStop();
    
    return( !ackbit );
}


//----------------------------------------------------------------------------
//  I2C_ReadDev()
//
//      Send i2c address then read a single byte.
//----------------------------------------------------------------------------
unsigned char I2C_ReadDev( unsigned char devAddr )
{
    unsigned char i2cData, ackbit;
    
    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );
    i2cData = i2cRead( _i2cNAK );
    i2cStop();
     
    return( i2cData );
}


//----------------------------------------------------------------------------
//  I2C_ReadDevReg()
//
//      Send i2c address, register byte, then read a byte.
//----------------------------------------------------------------------------
unsigned char I2C_ReadDevReg( unsigned char devAddr, unsigned char reg)
{
    unsigned char i2cData, ackbit;

    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   
    ackbit = ( ackbit << 1 ) | i2cWrite( reg );   
        
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );
    wait( 400000 );
    i2cData = i2cRead( _i2cNAK );
    i2cStop();

    return( i2cData );
}

  
//----------------------------------------------------------------------------
//  I2C_ReadDevCmd()
//
//      Send i2c address, register byte, and command byte for register.
//      Then read a byte.
//----------------------------------------------------------------------------
unsigned char I2C_ReadDevCmd( unsigned char devAddr, unsigned char reg, unsigned char cmd )
{
    unsigned char i2cData, ackbit;
  
    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   
    ackbit = ( ackbit << 1 ) | i2cWrite( reg ); 
    ackbit = ( ackbit << 1 ) | i2cWrite( cmd ); 

    wait( 1600000 );            // 20mS delay (for Matrix Orbital LK-25) TODO: optimize this
  
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );     // repeat with read bit now set
    i2cData = i2cRead( _i2cNAK );
    i2cStop();
    
    return( i2cData );
}

   
//----------------------------------------------------------------------------
//  I2C_WriteDev()
//
//      Send i2c address, then write one data byte, return ACK bit.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDev( unsigned char devAddr, unsigned char i2cData)
{
    unsigned char ackbit;

    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData );
    i2cStop();
    
    return( ackbit );
}

//----------------------------------------------------------------------------
//  I2C_WriteDev2()
//
//      Send i2c address, then write two data bytes, return ACK bit.
//      The two data bytes are often register / value pairs, but may be what-
//      ever the protocol the device uses.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDev2( unsigned char devAddr, unsigned char i2cData1, unsigned char i2cData2 )
{
    unsigned char ackbit;
    
    ackbit = _i2cACK;
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData1 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData2 );
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  I2C_WriteDev3()
//
//      Send i2c address, then write three data bytes, return ACK bit.
//      The three data bytes are whatever the protocol the device uses.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDev3( unsigned char devAddr, unsigned char i2cData1, unsigned char i2cData2, unsigned char i2cData3 )
{
    unsigned char ackbit;
    
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData1 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData2 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData3 );
    i2cStop();
    
    return( ackbit );
}
  

//----------------------------------------------------------------------------
//  I2C_WriteDev4()
//
//      Send i2c address, then write four data bytes, return ACK bit.
//      The four data bytes are whatever the protocol the device uses.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDev4( unsigned char devAddr, unsigned char i2cData1, unsigned char i2cData2, unsigned char i2cData3, unsigned char i2cData4 )
{
    unsigned char ackbit;
    
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData1 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData2 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData3 );
    ackbit = ( ackbit << 1 ) | i2cWrite( i2cData4 );
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  I2C_WriteDevBuf()
//
//      Send i2c address, then the number of bytes specified by the first byte
//      in the buffer. Returns the ACK bit from the device.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDevBuf( unsigned char devAddr, unsigned char *bufAddr )
{
    unsigned char ackbit, size, idx;
  
    size = *bufAddr++;          // get number of bytes to write
  
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    for ( idx = 0; idx < size; idx++ )
        ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  I2C_WriteDevRegBuf()
//
//      Send i2c address, the register to write to, then the number of bytes 
//      from the specified butter by the first byte in the buffer. Returns the
//      ACK bit from the device (i.e. 1st byte = # of bytes to write).
//----------------------------------------------------------------------------
unsigned char I2C_WriteDevRegBuf( unsigned char devAddr, unsigned char reg, unsigned char *bufAddr )
{
    unsigned char ackbit, size, idx;
  
    size = *bufAddr++;          // get number of bytes to write
  
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( reg );
    for ( idx = 0; idx < size; idx++ )
        ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  I2C_WriteDevStr()
//
//      Send i2c address, then the null terminated ASCII string pointed to by
//      the buffer address passed. Returns the ACK bit from the device.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDevStr( unsigned char devAddr, unsigned char *bufAddr )
{
    unsigned char ackbit;
    
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = (ackbit << 1) | i2cWrite(devAddr);
    while ( *bufAddr )
        ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );          
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  I2C_WriteDevRegStr()
//
//      Send i2c address, the register, then the null terminated ASCII string
//      pointed to by the buffer address passed. Returns the ACK bit from the
//      device.
//----------------------------------------------------------------------------
unsigned char I2C_WriteDevRegStr( unsigned char devAddr, unsigned char reg, unsigned char *bufAddr )
{
    unsigned char ackbit;
    
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = (ackbit << 1) | i2cWrite(devAddr);
    ackbit = (ackbit << 1) | i2cWrite(reg);
    while ( *bufAddr )
        ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );          
    i2cStop();
    
    return( ackbit );
}


//----------------------------------------------------------------------------
//  i2cStop()
//----------------------------------------------------------------------------
static void i2cStop( void )
{
    DIRA &= ~i2cClk;
    DIRA &= ~i2cDat;
}

    
//----------------------------------------------------------------------------
//  i2cStart()
//----------------------------------------------------------------------------
static void i2cStart( void )
{
    DIRA &= ~i2cDat;   
    DIRA &= ~i2cClk;
    DIRA |= i2cDat;       
       
    for(;;)
    {
        if ( INA & i2cClk ) break;
    }
}
      

//----------------------------------------------------------------------------
//  i2cWrite()
//
//      write a bite
//----------------------------------------------------------------------------
static unsigned char i2cWrite( unsigned char dat )
{
    unsigned char ackbit, idx, bit;
                                     
    DIRA |= i2cClk;             // init the clock line
 
    bit = 0x80;
 
    for( idx = 0; idx < 8; idx++ )
    { 
        if ( ( bit & dat ) == 0 )       // set the SDA while the SCL is LOW
            DIRA |= i2cDat;
        else
            DIRA &= ~i2cDat;
                       
        DIRA &= ~i2cClk;                // toggle SCL HIGH
        bit >>= 1;                      // slight delay too ...
        DIRA |= i2cClk;                 // toogle SCL LOW        
    }
       
    DIRA &= ~i2cDat;                    // setup for ACK - pin to input
                  
    DIRA &= ~i2cClk;                    // read in the ACK
    ackbit = ( ( INA & i2cDat ) != 0 );
    DIRA |= i2cClk;
     
    DIRA |= i2cDat;                     // leave the SDA pin LOW

    return( ackbit );  
}

    
//----------------------------------------------------------------------------
//  i2cRead()
//
//      read a byte
//----------------------------------------------------------------------------
static unsigned char i2cRead( unsigned char ackbit )
{
    unsigned char i2cData, idx;
    
    DIRA &= ~i2cDat;                    // set the SCL to output and the SDA to input
    DIRA |= i2cClk;

    i2cData = 0;                        // clock in the byte
    
    for ( idx = 0; idx < 8; idx++ )
    {
        i2cData <<= 1;
        DIRA &= ~i2cClk;
        if ( INA & i2cDat )
            i2cData |= 1;
        DIRA |= i2cClk;
    }
          
    DIRA |= i2cDat;                     // send the ACK or NAK
    DIRA &= ~i2cClk;
    if ( ackbit )
        DIRA &= ~i2cDat;
    else
        DIRA |= i2cDat;
        
    DIRA |= i2cClk;
    
    return( i2cData );                  // return the data
}


//============================================================================
//                        TERMS OF USE: MIT License                                                       
//============================================================================
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
//                                       
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
//============================================================================
         