/* =========================================================================
 * LTC1860 12-bit analog digital converter driver 
 * -------------------------------------------------------------------------
 * File:    ltc1860.c
 * Author:  Pavel Pinkas, Ph. D.
 * Version: 001 (released 2009-08-06)
 *
 * Copyright (c) 2009 Pavel Pinkas, Ph. D. <pavelp@dnastar.com>
 * See end of file for terms of use. 
 * -------------------------------------------------------------------------
 * CHANGELOG:
 * 2009/08/09: V002 - 14% speedup against V001
 * 2009/08/06: V001 - Initial Release   
 * ========================================================================= */

#include <propeller.h>

#include "ltc1860.h"

/* -------------------------------------------------------------------------
 * ltc1860_setup () - setup the device (automatically calls ltc1860_reset to 
 * initialize everything)  
 * ------------------------------------------------------------------------- */
void ltc1860_setup (
	LTC1860 *dev,
    unsigned short convPin,
    unsigned short sckPin,
    unsigned short sdoPin)
{
	// assign pins
	dev->SDO = sdoPin;
    dev->SCK = sckPin;
    dev->CONV = convPin;
    
    // pre-calculate the reading mask
    dev->inner.mask = (1 << dev->SDO);
    
    // precalculate TCONV time 
    dev->inner.tconv_cycles = TCONV_DURATION * CLKFREQ / 1000000;
    
    // default conversion settings 
    dev->inner.max = 0.0;
    dev->inner.min = 0.0;
    dev->inner.ctype = NO_CONVERSION;
    dev->inner.cf = (LTC1860_Converter) NULL;
    
    // initialize
    ltc1860_reset (dev);
}

/* -------------------------------------------------------------------------
 * ltc1860_setup () - initializes/resets the pins   
 * ------------------------------------------------------------------------- */
void ltc1860_reset (LTC1860 *dev)
{
	// SDO pin is input 
    CLR (DIRA, dev->SDO);
    
    // Clock and conversion pins are outputs
    SET (DIRA, dev->SCK);
    SET (DIRA, dev->CONV);
    
    // clock idles high, conversion idles low 
    SET (OUTA, dev->SCK);
    CLR (OUTA, dev->CONV);   
}
 
/* -------------------------------------------------------------------------
 * ltc1860_read () - make one reading   
 * ------------------------------------------------------------------------- */
unsigned short ltc1860_read (LTC1860 *dev)
{
	unsigned short iBit;
    unsigned long result = 0;
    long mask = dev->inner.mask;
    long wait = dev->inner.tconv_cycles;
    
    SET (OUTA, dev->CONV);	// start conversion 
    WAIT (wait);
    CLR (OUTA, dev->CONV);
        
    // read the result (values shifted out on the falling edge of the clock)
    for (iBit = 1; iBit <= CONVERTER_RESOLUTION; iBit++)
    {
    	result = result << 1; // shift existing data one position left 
        
        CLR (OUTA, dev->SCK);   // clock goes low to shift out next bit  
        result |= ((INA & mask) != 0); // read the next bit         
        SET (OUTA, dev->SCK);   // put clock back high 
    } 
    
    return result;
}

/* -------------------------------------------------------------------------
 * ltc1860_set_range () - set conversion range   
 * ------------------------------------------------------------------------- */
void ltc1860_set_range (
	LTC1860 *dev,
    float min,
    float max)
{
	dev->inner.min = min;
    dev->inner.max = max;
    dev->inner.ctype = RANGE_CONVERSION;
}

/* -------------------------------------------------------------------------
 * ltc1860_set_converter () - set custom converter function   
 * ------------------------------------------------------------------------- */
void ltc1860_set_converter (
	LTC1860 *dev,
    LTC1860_Converter f)
{
	dev->inner.cf = f;
    dev->inner.ctype = CUSTOM_CONVERTER;
}

/* -------------------------------------------------------------------------
 * ltc1860_set_raw () - removes converter or range conversion  
 * ------------------------------------------------------------------------- */
void ltc1860_set_raw (LTC1860 *dev)
{
	dev->inner.ctype = NO_CONVERSION;
    dev->inner.cf = (LTC1860_Converter) NULL;
}

/* -------------------------------------------------------------------------
 * ltc1860_convert () - make one reading and convert it   
 * ------------------------------------------------------------------------- */
float ltc1860_convert (LTC1860 *dev)
{
	unsigned short raw = ltc1860_read (dev);
    float result;
    
	switch (dev->inner.ctype)
    {
    	case NO_CONVERSION:
        	result = raw;
        	break;
            
        case RANGE_CONVERSION:
            result = 
            	     (dev->inner.max - dev->inner.min) 
                   * (float) raw / (float) CONVERTER_RANGE 
                   + dev->inner.min;
        	break;
            
        case CUSTOM_CONVERTER:
        	result = dev->inner.cf (raw);
        	break;
    }
    
    return result;
}

/* ===========================================================================
 *                         TERMS OF USE: MIT License
 * ---------------------------------------------------------------------------
 * Copyright (c) 2009 Pavel Pinkas, Ph. D. <pavelp@dnastar.com>
 * ---------------------------------------------------------------------------
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 *
 * ======================================================================== */
 
