/* =========================================================================
 * LTC1860 12-bit analog digital converter --- Examples of Use 
 * -------------------------------------------------------------------------
 * File:    ltc1860_test.c
 * Author:  Pavel Pinkas, Ph. D.
 * Version: 001 (released 2009-08-06)
 *
 * Copyright (c) 2009 Pavel Pinkas, Ph. D. <pavelp@dnastar.com>
 * See end of file for terms of use. 
 * -------------------------------------------------------------------------
 * CHANGELOG:
 * 2009/08/09: V002 - 14% speedup against V001 (no changes in this file)
 * 2009/08/06: V001 - Initial Release  
 * ========================================================================= */

#include <propeller.h>
#include <stdio.h>
#include <asio.h>
#include <math.h>

#include "ltc1860.h"

/* ------------------------------------------------------------------------
 * Prototypes
 * ------------------------------------------------------------------------ */
void main (void);
void example1 (void);
void example2 (void);
void example3 (void);

float pressure_altitude (unsigned short);

/* ------------------------------------------------------------------------
 * Serial terminal baud rate
 * ------------------------------------------------------------------------ */
#define BAUD_RATE 57600

/* ------------------------------------------------------------------------
 * Device configuration (change to fit your circuit)
 * ------------------------------------------------------------------------ */
#define CONV_PIN 16
#define SDO_PIN  17
#define SCK_PIN  18

/* ------------------------------------------------------------------------
 * main ()
 * ------------------------------------------------------------------------ */
void main (void)
{
	asio_init (BAUD_RATE);  // initialize serial terminal
    printf ("\n\n\n");
    
	example1 ();            // run example #1 
    example2 ();            // run example #2 
    example3 ();            // run example #3
}

/* ------------------------------------------------------------------------
 * Numeric constants used in all examples 
 * ------------------------------------------------------------------------ */
#define NUM_READINGS 10

/* ------------------------------------------------------------------------
 * example1 () --- shows how to set up the driver, take 10 consecutive 
 * readings and display the raw values (0 .. 4095) with no conversion 
 * ------------------------------------------------------------------------ */
void example1 (void)
{
	LTC1860 dev;
    int iReading;
    
    printf ("Example 1: raw readings, no data conversion\n");
    
    // setup and initialize the device 
    ltc1860_setup (&dev, CONV_PIN, SCK_PIN, SDO_PIN);
    
    // take and display the readings
    for (iReading = 1; iReading <= NUM_READINGS; iReading++)
    {
    	unsigned short reading = ltc1860_read (&dev);
        printf ("#%02d: 0x%03X\n", iReading, reading);	
    }
    
    printf ("\n");
}

/* ------------------------------------------------------------------------
 * Numeric constants used in examples 2 and 3 
 * ------------------------------------------------------------------------ */
#define MIN_VOLTAGE 0.0
#define MAX_VOLTAGE 5.0

/* ------------------------------------------------------------------------
 * example2 () --- shows how to set up the driver, take 10 consecutive  
 * readings and displayed measured voltages (conversion done via MIN/MAX 
 * settings)  
 * ------------------------------------------------------------------------ */
void example2 (void)
{
	LTC1860 dev;
    int iReading;
    
    printf ("Example 2: using range conversion\n");
    
    // setup and initialize the device and the data conversion
    ltc1860_setup (&dev, CONV_PIN, SCK_PIN, SDO_PIN);
    ltc1860_set_range (&dev, MIN_VOLTAGE, MAX_VOLTAGE);
    
    // take readings and displayed measured voltages 
    for (iReading = 1; iReading <= NUM_READINGS; iReading++)
    {
    	float reading = ltc1860_convert (&dev);
        printf ("#%02d: %6.4f [V]\n", iReading, reading);
    }
    
    printf ("\n");
}

/* ------------------------------------------------------------------------
 * Additional numeric constants for example 3 
 * ------------------------------------------------------------------------ */
#define REF_VOLTAGE 5.0

#define MIN_PRESSURE 15.0
#define MAX_PRESSURE 115.0

#define SEA_LEVEL_PRESSURE 101.325
#define PALT_EXPONENT 0.19025515825
#define PALT_COEFFICIENT (288.15 / 0.001990740740)

/* ------------------------------------------------------------------------
 * example3 () --- shows how to set up the driver, take 10 consecutive  
 * readings, use custom conversion function to convert raw data into 
 * measured quantity (altitude) and display the conversion results 
 * ------------------------------------------------------------------------ */
void example3 (void)
{
	LTC1860 dev;
    int iReading;
    
    printf ("Example 3: using custom converter\n");
    
    // setup and initialize the device and the data conversion
    ltc1860_setup (&dev, CONV_PIN, SCK_PIN, SDO_PIN);
    ltc1860_set_converter (&dev, pressure_altitude);
    
    // take readings and display measured altitude 
    for (iReading = 1; iReading <= NUM_READINGS; iReading++)
    {
    	float reading = ltc1860_convert (&dev);
        printf ("#%02d: %9.3f [ft]\n", iReading, reading);
    }
    
    printf ("\n");
}

/* ------------------------------------------------------------------------
 * pressure_altitude () --- custom conversion function that takes raw 
 * reading of the voltage from the pressure sensor and converts it to 
 * altitude (using a lookup table and linear interpolation) 
 * ------------------------------------------------------------------------ */

#define PRESS_TABLE_LENGTH 55
#define LOWEST_ALTITUDE -4000
#define ALTITUDE_STEP 1000

static float PressTable[PRESS_TABLE_LENGTH] = // lookup table
{
	116.855, 112.803, 108.866, 105.042, 101.327, 97.718, 94.214, 90.813, 87.512, 
     84.309,  81.201,  78.187,  75.264,  72.430, 69.683, 67.021, 64.442, 61.944,  
     59.525,  57.183,  54.916,  52.723,  50.601, 48.548, 46.564, 44.646, 42.792, 
     41.001,  39.272,  37.602,  35.989,  34.434, 32.933, 31.486, 30.090, 28.745, 
     27.449,  26.201,  24.999,  23.843,  22.730, 21.659, 20.630, 19.640, 18.690,
     17.777,  16.900,  16.059,  15.252,  14.478, 13.736, 13.026, 12.345, 11.694, 
     11.070
};

float pressure_altitude (unsigned short raw)
{
	float voltage;
    float pressure;
    float altitude;
	
    unsigned short iData;
        
	// convert raw reading into voltage
    voltage = REF_VOLTAGE * (float) raw / (float) CONVERTER_RANGE;
    
    // convert voltage into pressure
    pressure = 
               voltage / REF_VOLTAGE 
             * (MAX_PRESSURE - MIN_PRESSURE) 
             + MIN_PRESSURE;
             
    // find the closest pressure in the table
    for (iData = 1; iData < PRESS_TABLE_LENGTH; iData++)
    {
    	if (pressure >= PressTable[iData] && pressure <= PressTable[iData - 1])
        {
            // the pressure is somewhere between iData - 1 and iData
            // table entries 
        	break;
        }
    }
    
    if (iData < PRESS_TABLE_LENGTH)
    {
        // compute the whole-step portion of the altitude
        altitude = LOWEST_ALTITUDE + (iData - 1) * ALTITUDE_STEP;
        
        // use linear interpolation to improve the result 
        altitude += 
                    (pressure - PressTable[iData])
                  / (PressTable[iData - 1] - PressTable[iData])
                  * ALTITUDE_STEP;
    }
              
    return altitude;
}

/* ===========================================================================
 *                         TERMS OF USE: MIT License
 * ---------------------------------------------------------------------------
 * Copyright (c) 2009 Pavel Pinkas, Ph. D. <pavelp@dnastar.com>
 * ---------------------------------------------------------------------------
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 *
 * ======================================================================== */
 
